@@ -1,7 +1,7 @@
 #
 #   plot.im.R
 #
-#  $Revision: 1.158 $   $Date: 2024/02/04 08:04:51 $
+#  $Revision: 1.169 $   $Date: 2025/04/22 01:31:45 $
 #
 #  Plotting code for pixel images
 #
@@ -47,6 +47,8 @@ plot.im <- local({
       }
     }
     extrargs <- setdiff(extrargs, c("claim.title.space", "box"))
+    if(!is.na(k <- match("adj.main", names(aarg))))
+      names(aarg)[k] <- "adj"
     z <- do.call.matched(image.default,
                          append(imagedata, aarg),
                          extrargs=extrargs)
@@ -212,7 +214,7 @@ plot.im <- local({
     return(y)
   }
 
-  Ticks <- function(usr, log=FALSE, nint=NULL, ..., clip=TRUE) {
+  Ticks <- function(usr, log=FALSE, nint=NULL, ..., clip=TRUE, deco=identity) {
     #' modification of grDevices::axisTicks
     #'      constrains ticks to be inside the specified range if clip=TRUE
     #'      accepts nint=NULL as if it were missing
@@ -222,7 +224,8 @@ plot.im <- local({
       zlimits <- if(log) 10^usr else usr
       z <- z[inside.range(z, zlimits)]
     }
-    return(z)
+    if(!log) z <- deco(z)
+    return(unique(z))
   }
 
   numericalRange <- function(x, zlim=NULL) {
@@ -243,15 +246,18 @@ plot.im <- local({
   PlotIm <- function(x, ...,
                      main, 
                      add=FALSE, clipwin=NULL,
-                     col=NULL, valuesAreColours=NULL, log=FALSE,
+                     col=NULL, reverse.col=FALSE,
+                     valuesAreColours=NULL, log=FALSE,
                      ncolours=256, gamma=1, 
                      ribbon=show.all, show.all=!add,
+                     drop.ribbon=FALSE,
                      ribside=c("right", "left", "bottom", "top"),
                      ribsep=0.15, ribwid=0.05, ribn=1024,
                      ribscale=1, ribargs=list(), riblab=NULL, colargs=list(),
                      useRaster=NULL, workaround=FALSE, zap=1,
                      do.plot=TRUE,
-                     addcontour=FALSE, contourargs=list()) {
+                     addcontour=FALSE, contourargs=list(),
+                     background=NULL, clip.background=FALSE) {
     if(missing(main)) main <- short.deparse(substitute(x))
     verifyclass(x, "im")
     if(x$type == "complex") {
@@ -288,6 +294,9 @@ plot.im <- local({
       stop(paste("Log transform is undefined for an image of type",
                  sQuote(xtype)))
 
+    ## secret interface for handling log-transformed data
+    already.log <- identical(log, "already")
+    
     # determine whether pixel values are to be treated as colours
     if(!is.null(valuesAreColours)) {
       # argument given - validate
@@ -364,17 +373,27 @@ plot.im <- local({
                         "omitted from logarithmic colour map;",
                         "range of values =", prange(rx)),
                   call.=FALSE)
-        if(do.plot && !all(rx < 0))
+        if(do.plot && !all(rx > 0))
           warning("Zero pixel values omitted from logarithmic colour map",
                   call.=FALSE)
         x <- eval.im(log10orNA(x))
       } 
       xtype <- x$type
+      values.are.log <- TRUE
+      ## functions 'Log' and 'Exp' are used to determine tick marks and labels
       Log <- log10
-      Exp <- function(x) { 10^x }
+      Exp <- TenPower
+      if(!is.null(zlim))
+        dotargs$zlim <- zlim <- log10(zlim)
+    } else if(already.log) {
+      values.are.log <- TRUE
+      Log <- log10
+      Exp <- TenPower
     } else {
-      Log <- Exp <- function(x) { x }
+      values.are.log <- FALSE
+      Log <- Exp <- identity
     }
+    compress <- decompress <- identity
     
     imagebreaks <- NULL
 #    ribbonvalues <- ribbonbreaks <- NULL
@@ -410,86 +429,109 @@ plot.im <- local({
              if(!is.null(colmap)) {
                # explicit colour map
                s <- summary(colmap)
+               col <- s$outputs
                if(s$discrete)
                  stop("Discrete colour map is not applicable to real values")
                imagebreaks <- s$breaks
+               if(is.function(s$compress)) {
+                 ## remember these transformations
+                 compress <- s$compress
+                 decompress <- s$decompress
+                 ## transform pixel values to the compressed scale
+                 x <- eval.im(compress(x))
+                 imagebreaks <- compress(imagebreaks)
+                 values.are.log <- samefunction(compress, log10)
+               }
                vrange <- range(imagebreaks)
-               col <- s$outputs
-             } 
+             }
              trivial <- (diff(vrange) <= zap * .Machine$double.eps)
-             #' ribbonvalues: domain of colour map (pixel values)
-             #' ribbonrange: (min, max) of pixel values in image
-             #' nominalrange: range of values shown on ribbon 
-             #' nominalmarks: values shown on ribbon at tick marks
-             #' ribbonticks: pixel values of tick marks 
+             #' ribbonvalues: a sequence of pixel values, mapped to colours
+             #' ribbonrange:  (min, max) of pixel values mapped by ribbon
+             #' nominalrange: range of (scaled) values shown on ribbon 
+             #' nominalmarks: (scaled) values shown on ribbon at tick marks
+             #' ribbonticks: pixel values corresponding to tick marks 
              #' ribbonlabels: text displayed at tick marks
-             ribbonvalues <- if(trivial) vrange[1] else
-                             seq(from=vrange[1], to=vrange[2], length.out=ribn)
-             ribbonrange <- vrange
-             nominalrange <- Log(ribscale * Exp(ribbonrange))
-             nominalmarks <- if(trivial) nominalrange[1] else
-                             (user.ticks %orifnull% Ticks(nominalrange,
-                                                         log=do.log,
-                                                         nint=user.nint))
-             ribbonticks <- Log(nominalmarks/ribscale)
+             #' reusableticks: reusable value of user.ticks
+             if(trivial) {
+               ribbonvalues <- mean(vrange)
+               nominalmarks <- compress(Log(ribscale * Exp(decompress(ribbonvalues))))
+             } else {
+               ribbonvalues <- seq(from=vrange[1L], to=vrange[2L],
+                                   length.out=ribn)
+               ribbonrange <- vrange
+               nominalrange <- compress(Log(ribscale * Exp(decompress(ribbonrange))))
+               nominalmarks <- user.ticks %orifnull% Ticks(nominalrange,
+                                                           log=values.are.log,
+                                                           nint=user.nint,
+                                                           deco=decompress)
+             }
+             ribbonticks <- compress(Log(nominalmarks/ribscale))
              ribbonlabels <- user.ribbonlabels %orifnull% paste(nominalmarks)
+             reusableticks <- nominalmarks
            },
            integer = {
-             values <- as.vector(x$v)
-             values <- values[!is.na(values)]
-             uv <- unique(values)
-             vrange <- numericalRange(uv, zlim)
-             nvalues <- length(uv)
-             trivial <- (nvalues < 2)
-             if(!trivial){
-               nominalrange <- Log(ribscale * Exp(vrange))
-               if(!is.null(user.ticks)) {
-                 nominalmarks <- user.ticks
-               } else {
-                 nominalmarks <- Ticks(nominalrange,
-                                       log=do.log,
-                                       nint = user.nint)
-                 nominalmarks <- nominalmarks[nominalmarks %% 1 == 0]
-               }
-               ribbonticks <- Log(nominalmarks/ribscale)
-               ribbonlabels <- user.ribbonlabels %orifnull% paste(nominalmarks)
-               if(!do.log && isTRUE(all.equal(ribbonticks,
-                                              vrange[1]:vrange[2]))) {
-                 # each possible pixel value will appear in ribbon
-                 ribbonvalues <- vrange[1]:vrange[2]
-                 imagebreaks <- c(ribbonvalues - 0.5, vrange[2] + 0.5)
-                 ribbonrange <- range(imagebreaks)
-                 ribbonticks <- ribbonvalues
-                 ribbonlabels <- user.ribbonlabels %orifnull% paste(ribbonticks * ribscale)
-               } else {
-                 # not all possible values will appear in ribbon
-                 ribn <- min(ribn, diff(vrange)+1)
-                 ribbonvalues <- seq(from=vrange[1], to=vrange[2],
-                                     length.out=ribn)
-                 ribbonrange <- vrange
-               }
-             }
+             vrange <- numericalRange(x, zlim)
              if(!is.null(colmap)) {
                # explicit colour map
                s <- summary(colmap)
-               imagebreaks <-
-                 if(!s$discrete) s$breaks else
-                 c(s$inputs[1] - 0.5, s$inputs + 0.5)
                col <- s$outputs
+               if(s$discrete) {
+                 imagebreaks <- c(s$inputs[1] - 0.5, s$inputs + 0.5)
+               } else {
+                 imagebreaks <- s$breaks
+                 if(is.function(s$compress)) {
+                   ## remember these transformations
+                   compress <- s$compress
+                   decompress <- s$decompress
+                   ## transform pixel values to the compressed scale
+                   x <- eval.im(compress(x))
+                   imagebreaks <- compress(imagebreaks)
+                   vrange <- range(imagebreaks)
+                   values.are.log <- samefunction(compress, log10)
+                 }
+               }
+             } 
+             trivial <- (diff(vrange) < sqrt(.Machine$double.eps))
+             nominalrange <- Log(ribscale * Exp(vrange))
+             if(!is.null(user.ticks)) {
+               nominalmarks <- user.ticks
+             } else {
+               nominalmarks <- Ticks(nominalrange,
+                                     log=do.log,
+                                     nint = user.nint,
+                                     deco = decompress)
+               nominalmarks <- nominalmarks[nominalmarks %% 1 == 0]
+               nominalmarks <- decompress(nominalmarks)
+             }
+             reusableticks <- nominalmarks
+             ribbonticks <- compress(Log(nominalmarks/ribscale))
+             ribbonlabels <- user.ribbonlabels %orifnull% paste(nominalmarks)
+             if(!do.log && isTRUE(all.equal(ribbonticks,
+                                            vrange[1]:vrange[2]))) {
+               #' each possible pixel value will appear in ribbon
+               ribbonvalues <- vrange[1]:vrange[2]
+               imagebreaks <- c(ribbonvalues - 0.5, vrange[2] + 0.5)
+               ribbonrange <- range(imagebreaks)
+               ribbonticks <- ribbonvalues
+               ribbonlabels <- user.ribbonlabels %orifnull% paste(ribbonticks * ribscale)
+             } else {
+               ## not all possible values will appear in ribbon
+               ribn <- min(ribn, diff(vrange)+1)
+               ribbonvalues <- seq(from=vrange[1], to=vrange[2],
+                                   length.out=ribn)
+               ribbonrange <- vrange
              }
            },
            logical = {
-             values <- as.integer(as.vector(x$v))
-             values <- values[!is.na(values)]
-             uv <- unique(values)
-             trivial <- (length(uv) < 2)
              vrange <- c(0,1)
+             trivial <- FALSE
              imagebreaks <- c(-0.5, 0.5, 1.5)
              ribbonvalues <- c(0,1)
              ribbonrange <- range(imagebreaks)
 #             ribbonbreaks <- imagebreaks
              ribbonticks <- user.ticks %orifnull% ribbonvalues
              ribbonlabels <- user.ribbonlabels %orifnull% c("FALSE", "TRUE")
+             reusableticks <- ribbonticks
              if(!is.null(colmap)) 
                col <- colmap(c(FALSE,TRUE))
            },
@@ -506,6 +548,7 @@ plot.im <- local({
 #             ribbonbreaks <- imagebreaks
              ribbonticks <- user.ticks %orifnull% ribbonvalues
              ribbonlabels <- user.ribbonlabels %orifnull% paste(lev)
+             reusableticks <- ribbonticks
              vrange <- range(intlev)
              if(!is.null(colmap) && !valuesAreColours) 
                col <- colmap(fac)
@@ -524,6 +567,7 @@ plot.im <- local({
 #             ribbonbreaks <- imagebreaks
              ribbonticks <- user.ticks %orifnull% ribbonvalues
              ribbonlabels <- user.ribbonlabels %orifnull% paste(lev)
+             reusableticks <- ribbonticks
              vrange <- range(intlev)
              if(!is.null(colmap)) 
                col <- colmap(fac)
@@ -562,6 +606,11 @@ plot.im <- local({
       ## transform to grey scale
       colourinfo$col <- to.grey(colourinfo$col)
     }
+
+    if(isTRUE(reverse.col) && !valuesAreColours) {
+      ## reverse the colour sequence (using rev.colourmap or rev.default)
+      colourinfo$col <- rev(colourinfo$col)
+    }
     
     # colour map to be returned (invisibly)
     i.col <- colourinfo$col
@@ -569,13 +618,30 @@ plot.im <- local({
     output.colmap <-
       if(is.null(i.col)) NULL else
       if(inherits(i.col, "colourmap")) i.col else
+      if(inherits(colmap, "colourmap")) colmap else
       if(valuesAreColours) colourmap(col=i.col, inputs=i.col) else
       switch(xtype,
              integer=,
              real= {
-               if(!is.null(i.bks)) {
-                 colourmap(col=i.col, breaks=i.bks)
-               } else colourmap(col=i.col, range=vrange, gamma=gamma)
+               if(!do.log) {
+                 ## linear colour map 
+                 if(!is.null(i.bks)) {
+                   ## possibly uneven breaks
+                   colourmap(col=i.col, breaks=i.bks)
+                 } else {
+                   colourmap(col=i.col, range=vrange, gamma=gamma)
+                 }
+               } else {
+                 ## logarithmic colour map
+                 if(!is.null(i.bks)) {
+                   colourmap(col=i.col, breaks=TenPower(i.bks),
+                             compress=log10, decompress=TenPower)
+                 } else {
+                   colourmap(col=i.col, range=TenPower(vrange),
+                             gamma=gamma,
+                             compress=log10, decompress=TenPower)
+                 }
+               }
              },
              logical={
                colourmap(col=i.col, inputs=c(FALSE,TRUE))
@@ -586,6 +652,9 @@ plot.im <- local({
              },
              NULL)
 
+    ## save tickmark values
+    attr(output.colmap, "at") <- reusableticks
+    
     ## gamma correction
     soc <- summary(output.colmap)
     if(!is.null(gamma <- soc$gamma) && gamma != 1)
@@ -617,16 +686,38 @@ plot.im <- local({
 
     ## ........ catch old usage (undocumented ) ................
     contourargs <- resolve.defaults(contourargs, dotargs$args.contour)
+
+    ## ........ background object ..............................
+    if(!is.null(background)) {
+      if(isTRUE(clip.background)) {
+        bkg <- try(background[as.rectangle(x), drop=FALSE], silent=TRUE)
+        if(inherits(bkg, "try-error")) {
+          warning("Unable to clip the background object", call.=FALSE)
+        } else {
+          background <- bkg
+        }
+      }
+      backbox <- Frame(background)
+    } else {
+      backbox <- NULL
+    }
     
     ## ........ start plotting .................
 
-    if(!identical(ribbon, TRUE) || trivial) {
+    if(!isTRUE(ribbon) || (trivial && isTRUE(drop.ribbon))) {
       ## no ribbon wanted
 
-      attr(output.colmap, "bbox") <- as.rectangle(x)
+      attr(output.colmap, "bbox") <- boundingbox(as.rectangle(x), backbox)
       if(!do.plot)
         return(output.colmap)
 
+      ## plot background if specified
+      if(!is.null(background)) {
+        plot(background, main=main)
+        add <- TRUE
+        main <- ""
+      }
+
       ## plot image without ribbon
       image.doit(imagedata=list(x=cellbreaks(x$xcol, x$xstep),
                                 y=cellbreaks(x$yrow, x$ystep),
@@ -642,7 +733,7 @@ plot.im <- local({
                  list(useRaster=useRaster),
                  colourinfo,
                  list(zlim=vrange, asp = 1, main = main),
-                 list(values.are.log=do.log))
+                 list(values.are.log=values.are.log))
 ##      if(add && show.all)
 ##        fakemaintitle(x, main, dotargs)
 
@@ -681,7 +772,7 @@ plot.im <- local({
                             bb$yrange[1] - c(ribsep+ribwid, ribsep) * Size)
              rib.iside <- 1
            })
-    bb.all <- boundingbox(bb.rib, bb)
+    bb.all <- boundingbox(bb.rib, bb, backbox)
 
     attr(output.colmap, "bbox") <- bb.all
     attr(output.colmap, "bbox.legend") <- bb.rib
@@ -712,6 +803,10 @@ plot.im <- local({
                       extrargs=graphicsPars("owin"))
       main <- ""
     }
+    if(!is.null(background)) {
+      ## plot background
+      plot(background, add=TRUE)
+    }
     # plot image
     image.doit(imagedata=list(x=cellbreaks(x$xcol, x$xstep),
                               y=cellbreaks(x$yrow, x$ystep),
@@ -727,7 +822,7 @@ plot.im <- local({
                list(useRaster=useRaster),
                colourinfo,
                list(zlim=vrange, asp = 1, main = main),
-               list(values.are.log=do.log))
+               list(values.are.log=values.are.log))
 
 ##    if(add && show.all)
 ##      fakemaintitle(bb.all, main, ...)
@@ -773,7 +868,7 @@ plot.im <- local({
                list(main="", sub="", xlab="", ylab=""),
                dotargs,
                colourinfo,
-               list(values.are.log=do.log))
+               list(values.are.log=values.are.log))
     # box around ribbon?
     resol <- resolve.defaults(ribargs, dotargs)
     if(!identical(resol$box, FALSE))
@@ -792,29 +887,45 @@ plot.im <- local({
       axisargs <- list(side=ribaxis.iside, labels=ribbonlabels)
       switch(ribside,
              right={
-               scal <- diff(bb.rib$yrange)/diff(ribbonrange)
-               at <- bb.rib$yrange[1] + scal * (ribbonticks - ribbonrange[1])
+               if(trivial) {
+                 at <- mean(bb.rib$yrange)
+               } else {
+                 scal <- diff(bb.rib$yrange)/diff(ribbonrange)
+                 at <- bb.rib$yrange[1] + scal * (ribbonticks - ribbonrange[1])
+               }
                axisargs <- append(axisargs, list(at=at))
                posargs <- list(pos=bb.rib$xrange[2],
                                yaxp=c(bb.rib$yrange, length(ribbonticks)))
              },
              left={
-               scal <- diff(bb.rib$yrange)/diff(ribbonrange)
-               at <- bb.rib$yrange[1] + scal * (ribbonticks - ribbonrange[1])
+               if(trivial) {
+                 at <- mean(bb.rib$yrange)
+               } else {
+                 scal <- diff(bb.rib$yrange)/diff(ribbonrange)
+                 at <- bb.rib$yrange[1] + scal * (ribbonticks - ribbonrange[1])
+               }
                axisargs <- append(axisargs, list(at=at))
                posargs <- list(pos=bb.rib$xrange[1],
                                yaxp=c(bb.rib$yrange, length(ribbonticks)))
              },
              top={
-               scal <- diff(bb.rib$xrange)/diff(ribbonrange)
-               at <- bb.rib$xrange[1] + scal * (ribbonticks - ribbonrange[1])
+               if(trivial) {
+                 at <- mean(bb.rib$xrange)
+               } else {
+                 scal <- diff(bb.rib$xrange)/diff(ribbonrange)
+                 at <- bb.rib$xrange[1] + scal * (ribbonticks - ribbonrange[1])
+               }
                axisargs <- append(axisargs, list(at=at))
                posargs <- list(pos=bb.rib$yrange[2],
                                xaxp=c(bb.rib$xrange, length(ribbonticks)))
              },
              bottom={
-               scal <- diff(bb.rib$xrange)/diff(ribbonrange)
-               at <- bb.rib$xrange[1] + scal * (ribbonticks - ribbonrange[1])
+               if(trivial) {
+                 at <- mean(bb.rib$xrange)
+               } else {
+                 scal <- diff(bb.rib$xrange)/diff(ribbonrange)
+                 at <- bb.rib$xrange[1] + scal * (ribbonticks - ribbonrange[1])
+               }
                axisargs <- append(axisargs, list(at=at))
                posargs <- list(pos=bb.rib$yrange[1],
                                xaxp=c(bb.rib$xrange, length(ribbonticks)))
@@ -1036,8 +1147,11 @@ contour.im <- function (x, ..., main, axes=FALSE, add=FALSE,
   return(invisible(result))
 }
 
+
 ## not exported:
 
+TenPower <- function(x) { 10^x }
+
 log10orNA <- function(x) {
   y <- rep(NA_real_, length(x))
   ok <- !is.na(x) & (x > 0)
